package dev.fyloz.colorrecipesexplorer.service.files

import dev.fyloz.colorrecipesexplorer.JavaFile
import dev.fyloz.colorrecipesexplorer.utils.File
import dev.fyloz.colorrecipesexplorer.utils.FilePath
import mu.KotlinLogging

object FileCache {
    private val logger = KotlinLogging.logger {}
    private val cache = hashMapOf<String, CachedFileSystemItem>()

    operator fun contains(filePath: FilePath) =
        filePath.path in cache

    operator fun get(filePath: FilePath) =
        cache[filePath.path]

    fun getDirectory(filePath: FilePath) =
        if (directoryExists(filePath)) {
            this[filePath] as CachedDirectory
        } else {
            null
        }

    fun getFile(filePath: FilePath) =
        if (fileExists(filePath)) {
            this[filePath] as CachedFile
        } else {
            null
        }

    private operator fun set(filePath: FilePath, item: CachedFileSystemItem) {
        cache[filePath.path] = item
    }

    fun exists(filePath: FilePath) =
        filePath in this && cache[filePath.path]!!.exists

    fun directoryExists(filePath: FilePath) =
        exists(filePath) && this[filePath] is CachedDirectory

    fun fileExists(filePath: FilePath) =
        exists(filePath) && this[filePath] is CachedFile

    fun setExists(filePath: FilePath, exists: Boolean = true) {
        if (filePath !in this) {
            load(filePath)
        }

        this[filePath] = this[filePath]!!.clone(exists)
        logger.debug("Updated FileCache state: ${filePath.path} exists -> $exists")
    }

    fun setDoesNotExists(filePath: FilePath) =
        setExists(filePath, false)

    fun load(filePath: FilePath) =
        with(JavaFile(filePath.path).toFileSystemItem()) {
            cache[filePath.path] = this

            logger.debug("Loaded file at ${filePath.path} into FileCache")
        }

    fun addContent(filePath: FilePath, childFilePath: FilePath) {
        val directory = prepareDirectory(filePath) ?: return

        val updatedContent = setOf(
            *directory.content.toTypedArray(),
            JavaFile(childFilePath.path).toFileSystemItem()
        )

        this[filePath] = directory.copy(content = updatedContent)
        logger.debug("Added child ${childFilePath.path} to ${filePath.path} in FileCache")
    }

    fun removeContent(filePath: FilePath, childFilePath: FilePath) {
        val directory = prepareDirectory(filePath) ?: return

        val updatedContent = directory.content
            .filter { it.path.path != childFilePath.path }
            .toSet()

        this[filePath] = directory.copy(content = updatedContent)
        logger.debug("Removed child ${childFilePath.path} from ${filePath.path} in FileCache")
    }

    private fun prepareDirectory(filePath: FilePath): CachedDirectory? {
        if (!directoryExists(filePath)) {
            logger.warn("Cannot add child to ${filePath.path} because it is not in the cache")
            return null
        }

        val directory = getDirectory(filePath)
        if (directory == null) {
            logger.warn("Cannot add child to ${filePath.path} because it is not a directory")
            return null
        }

        return directory
    }
}

interface CachedFileSystemItem {
    val name: String
    val path: FilePath
    val exists: Boolean

    fun clone(exists: Boolean): CachedFileSystemItem
}

data class CachedFile(
    override val name: String,
    override val path: FilePath,
    override val exists: Boolean
) : CachedFileSystemItem {
    constructor(file: File) : this(file.name, file.toFilePath(), file.exists() && file.isFile)

    override fun clone(exists: Boolean) =
        this.copy(exists = exists)
}

data class CachedDirectory(
    override val name: String,
    override val path: FilePath,
    override val exists: Boolean,
    val content: Set<CachedFileSystemItem> = setOf()
) : CachedFileSystemItem {
    constructor(file: File) : this(file.name, file.toFilePath(), file.exists() && file.isDirectory, file.fetchContent())

    val contentFiles: Collection<CachedFile>
        get() = content.filterIsInstance<CachedFile>()

    override fun clone(exists: Boolean) =
        this.copy(exists = exists)

    companion object {
        private fun File.fetchContent() =
            (this.file.listFiles() ?: arrayOf<JavaFile>())
                .filterNotNull()
                .map { it.toFileSystemItem() }
                .toSet()
    }
}

fun JavaFile.toFileSystemItem() =
    if (this.isDirectory) {
        CachedDirectory(File(this))
    } else {
        CachedFile(File(this))
    }